'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _es6Promisify = require('es6-promisify');

var _es6Promisify2 = _interopRequireDefault(_es6Promisify);

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _String = require('./String');

var _String2 = _interopRequireDefault(_String);

var _Domain = require('./Domain');

var _Domain2 = _interopRequireDefault(_Domain);

var _IP = require('./IP');

var _IP2 = _interopRequireDefault(_IP);

var _Error = require('../Error');

var _Error2 = _interopRequireDefault(_Error);

var _Reference = require('../Reference');

var _Reference2 = _interopRequireDefault(_Reference);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const debug = (0, _debug2.default)('validator:email');

let myIP = '';
function getMyIP() {
  if (myIP.length) return Promise.resolve(myIP);
  return Promise.resolve().then(() => require('request-promise-native')).then(request => request('http://ipinfo.io/ip')).then(html => new _IP2.default().validate(html.trim())).then(ip => {
    myIP = ip;
    return ip;
  }).catch(err => Promise.reject(new Error(`Could not get own IP address (needed for further checks): ${err.message}`)));
}

function connect(record) {
  const domain = record.exchange;
  return getMyIP().then(() => Promise.resolve().then(() => require('net'))).then(net => new Promise(resolve => {
    let timer;
    const client = net.createConnection(25, domain, () => {
      debug('Send HELO command to mailserver...');
      client.write(`HELO ${myIP}\r\n`);
      client.write('QUIT\r\n');
      client.end();
    });
    let res = '';
    client.on('data', data => {
      clearTimeout(timer);
      res += data.toString();
    });
    client.on('error', err => {
      clearTimeout(timer);
      debug(`Error from server ${domain}: ${err.message}`);
      resolve(false);
    });
    client.on('end', () => {
      clearTimeout(timer);
      if (res.length) {
        debug(`Server ${domain} responded with:\n${res.trim()}`);
        return resolve(true);
      }
      debug(`No valid response from server ${domain}`);
      return resolve(false);
    });
    timer = setTimeout(() => {
      debug('Attempt at connection exceeded timeout value');
      client.end();
      resolve(false);
    }, 5000);
  }));
}

function checklist(name, access, record) {
  const domain = record.exchange;
  const reverse = `${domain.split('.').reverse().join('.')}.${access.zone}`;
  return Promise.resolve().then(() => require('dns')).then(dns => (0, _es6Promisify2.default)(dns.resolve)(reverse)).then(() => name).catch(() => '');
}

class EmailSchema extends _String2.default {
  constructor(base) {
    super(base);

    let raw = this._rules.descriptor.pop();
    let allow = this._rules.descriptor.pop();
    this._rules.descriptor.push(this._structDescriptor, allow, this._connectDescriptor, this._blacklistDescriptor, this._greylistDescriptor, this._formatDescriptor, raw);
    raw = this._rules.validator.pop();
    allow = this._rules.validator.pop();
    this._rules.validator.push(this._structValidator, allow, this._connectValidator, this._blacklistValidator, this._greylistValidator, this._formatValidator, raw);
  }

  stripEmpty() {
    return this._setError('stripEmpty');
  }
  truncate() {
    return this._setError('truncate');
  }
  pad() {
    return this._setError('pad');
  }

  dns(flag) {
    return this._setFlag('dns', flag);
  }
  punycode(flag) {
    return this._setFlag('punycode', flag);
  }
  resolve(flag) {
    return this._setFlag('resolve', flag);
  }

  _typeDescriptor() {
    return 'It has to be a reasonable email address with optional descriptive part.\n';
  }

  normalize(flag) {
    return this._setFlag('normalize', flag);
  }

  _structDescriptor() {
    const set = this._setting;
    let msg = '';
    if (set.normalize) {
      if (this._isReference('normalize')) {
        msg += `Extended formats like additional domains, sub domains and tags which mostly belong to \
the same mailbox will be removed. if set under ${set.normalize.description}. `;
      } else {
        msg += 'Extended formats like additional domains, sub domains and tags which mostly belong \
to the same mailbox will be removed. ';
      }
    }
    const schema = new _Domain2.default();
    if (set.dns) schema.dns('MX');
    if (set.punycode) schema.punycode();else if (set.resolve) schema.resolve();
    return `${msg}- domain part: ${schema.description}\n\n`;
  }

  _structValidator(data) {
    const check = this._check;

    const match = data.value.match(/^(.*\S)\s+<(.*)>\s*$/);
    const full = (match ? match[2] : data.value).trim();
    const result = {};
    if (match) result.name = match[1];else result.name = null;
    const at = full.lastIndexOf('@');
    if (at === -1) result.local = full;else {
      result.local = full.substring(0, at);
      result.domain = full.substring(at + 1);
    }
    data.value = result;

    if (check.normalize) {
      if (result.domain.match(/^g(oogle)?mail\.com$/i)) {
        result.local = result.local.replace(/\.|\+.*$/g, '');
        result.domain = 'gmail.com';
      }
      if (result.domain.match(/^facebook\.com$/i)) {
        result.local = result.local.replace(/\./g, '');
      }
    }

    if (!data.value.domain) {
      return Promise.reject(new _Error2.default(this, data, 'The email address is missing the server \
part starting with \'@\''));
    }
    if (data.value.local.length > 64) {
      return Promise.reject(new _Error2.default(this, data, 'The local mailbox name is too long (64 \
chars max per specification)'));
    }

    const schema = new _Domain2.default();
    if (check.dns) schema.dns('MX');
    if (check.punycode) schema.punycode();else if (check.resolve) schema.resolve();else schema.raw();
    return schema._validate(data.sub('domain')).then(d => {
      data.value.domain = d.value;
      return Promise.resolve();
    });
  }

  _allowValidator(data) {
    const check = this._check;
    this._checkArrayString('allow');
    this._checkArrayString('deny');

    let denyPriority = 0;
    let allowPriority = 0;
    const email = `${data.value.local}@${data.value.domain || 'localhost'}`.toLowerCase();
    if (check.deny && check.deny.length) {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = check.deny[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          const e = _step.value;

          const match = e.match(/^(.*\S)\s+<(.*)>\s*$/);
          let full = (match ? match[2] : e).trim().toLowerCase();
          const at = full.lastIndexOf('@');
          let domain = null;
          if (at === -1) {
            domain = full;
            full = undefined;
          } else domain = full.substring(at + 1);
          if (email === full) {
            denyPriority = 99;
            break;
          }
          if (email.endsWith(`.${domain}`) || email.endsWith(`@${domain}`)) {
            const m = domain.match(/\./g);
            const level = m ? m.length + 1 : 1;
            if (level > denyPriority) denyPriority = level;
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }
    if (check.allow && check.allow.length) {
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = check.allow[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          const e = _step2.value;

          const match = e.match(/^(.*\S)\s+<(.*)>\s*$/);
          let full = (match ? match[2] : e).trim().toLowerCase();
          const at = full.lastIndexOf('@');
          let domain = null;
          if (at === -1) {
            domain = full;
            full = undefined;
          } else domain = full.substring(at + 1);
          if (email === full) {
            allowPriority = 99;
            break;
          }
          if (email.endsWith(`.${domain}`) || email.endsWith(`@${domain}`)) {
            const m = domain.match(/\./g);
            const level = m ? m.length + 1 : 1;
            if (level > allowPriority) allowPriority = level;
          }
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }
    }
    if (denyPriority > allowPriority) {
      return Promise.reject(new _Error2.default(this, data, 'Email found in blacklist (denied item).'));
    }
    return Promise.resolve();
  }

  connect(flag) {
    return this._setFlag('connect', flag);
  }

  _connectDescriptor() {
    const set = this._setting;
    let msg = '';
    if (set.connect) {
      if (this._isReference('connect')) {
        msg += `A handshake with the mail server should be possible possible if set under \
${set.connect.description}. `;
      } else msg += 'A handshake with the mail server should be possible. ';
    }
    return msg.length ? `${msg.trim()}\n` : msg;
  }

  _connectValidator(data) {
    const check = this._check;
    try {
      this._checkBoolean('connect');
    } catch (err) {
      return Promise.reject(new _Error2.default(this, data, err.message));
    }

    if (check.connect) {
      return Promise.resolve().then(() => require('dns')).then(dns => (0, _es6Promisify2.default)(dns.resolve)(data.value.domain, 'MX')).then(list => Promise.all(list.map(e => connect(e)))).then(res => {
        if (res.filter(e => e).length) return undefined;
        return Promise.reject(new Error('No correct responding mail server could be detected for this domain'));
      });
    }
    return Promise.resolve();
  }

  blackList(flag) {
    return this._setFlag('blackList', flag);
  }

  _blacklistDescriptor() {
    const set = this._setting;
    let msg = '';
    if (set.blackList) {
      if (this._isReference('blackList')) {
        msg += `The mail server should not be on any black list if set under \
${set.blackList.description}. `;
      } else msg += 'The mail server should not be on any black list. ';
    }
    return msg.length ? `${msg.trim()}\n` : msg;
  }

  _blacklistValidator(data) {
    const check = this._check;
    try {
      this._checkBoolean('blackList');
    } catch (err) {
      return Promise.reject(new _Error2.default(this, data, err.message));
    }

    if (check.blackList) {
      return Promise.resolve().then(() => require('../../config/blacklists.json')).then(list => Promise.all(Object.keys(list).map(k => {
        debug(`check in blacklist: ${k}`);
        return Promise.resolve().then(() => require('dns')).then(dns => (0, _es6Promisify2.default)(dns.resolve)(data.value.domain, 'MX')).then(addr => Promise.all(addr.map(e => checklist(k, list[k], e)))).then(res => {
          const found = res.filter(e => !e);
          if (found.length) return undefined;
          return Promise.reject(new Error(`Found on the blacklists: ${found.join(', ')}`));
        });
      }))).then();
    }
    return Promise.resolve();
  }

  greyList(flag) {
    return this._setFlag('greyList', flag);
  }

  _greylistDescriptor() {
    const set = this._setting;
    let msg = '';
    if (set.greyList) {
      if (this._isReference('greyList')) {
        msg += `The mail server should not be on any black list if set under \
${set.greyList.description}. `;
      } else msg += 'The mail server should not be on any black list. ';
    }
    return msg.length ? `${msg.trim()}\n` : msg;
  }

  _greylistValidator(data) {
    const check = this._check;
    try {
      this._checkBoolean('greyList');
    } catch (err) {
      return Promise.reject(new _Error2.default(this, data, err.message));
    }

    if (check.greyList) {
      return Promise.resolve().then(() => require('../../config/greylists.json')).then(list => Promise.all(Object.keys(list).map(k => {
        debug(`check in greylist: ${k}`);
        return Promise.resolve().then(() => require('dns')).then(dns => (0, _es6Promisify2.default)(dns.resolve)(data.value.domain, 'MX')).then(addr => Promise.all(addr.map(e => checklist(k, list[k], e)))).then(res => {
          const found = res.filter(e => !e);
          if (found.length) return undefined;
          return Promise.reject(new Error(`Found on the greylists: ${found.join(', ')}`));
        });
      }))).then();
    }
    return Promise.resolve();
  }

  withName(flag) {
    return this._setFlag('withName', flag);
  }

  _formatDescriptor() {
    const set = this._setting;
    let msg = '';
    if (set.withName) {
      if (this._isReference('withName')) {
        msg += `The email address may contain a descriptive name if defined under \
${set.withName.description}. `;
      } else msg += 'The email address may contain a descriptive name. ';
    }
    return msg.length ? `${msg.trim()}\n` : msg;
  }

  _formatValidator(data) {
    const check = this._check;
    try {
      this._checkBoolean('withName');
    } catch (err) {
      return Promise.reject(new _Error2.default(this, data, err.message));
    }

    let email = data.value.local;
    if (data.value.domain) email += `@${data.value.domain}`;
    if (check.withName && data.value.name) email = `${data.value.name} <${email}>`;
    data.value = email;
    return Promise.resolve();
  }
}

exports.default = EmailSchema;